from btypes import *

class FormatError(Exception): pass

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    unknown0 = uint16
    __padding__ = Padding(2)
    shape_packet_count = uint32
    position_count = uint32
    node_offset = uint32

    def __init__(self):
        self.magic = b'INF1'


class NodeType:
    END_GRAPH = 0x00
    BEGIN_CHILDREN = 0x01
    END_CHILDREN = 0x02
    JOINT = 0x10
    MATERIAL = 0x11
    SHAPE = 0x12


class Node(Struct):
    node_type = uint16
    index = uint16

    def __init__(self,node_type,index):
        self.node_type = node_type
        self.index = index


END_GRAPH_NODE = Node(NodeType.END_GRAPH,0)
BEGIN_CHILDREN_NODE = Node(NodeType.BEGIN_CHILDREN,0)
END_CHILDREN_NODE = Node(NodeType.END_CHILDREN,0)


class JointNode(Node):

    def __init__(self,index):
        super().__init__(NodeType.JOINT,index)


class MaterialNode(Node):

    def __init__(self,index):
        super().__init__(NodeType.MATERIAL,index)


class ShapeNode(Node):

    def __init__(self,index):
        super().__init__(NodeType.SHAPE,index)


class SceneGraph:

    def __init__(self):
        self.unknown0 = 0
        self.children = []

#------------------------------------------------------------------------------

def pack_children(stream,children):
    for node in children:
        Node.pack(stream,node)
        if hasattr(node,'children') and node.children:
            Node.pack(stream,BEGIN_CHILDREN_NODE)
            pack_children(stream,node.children)

    Node.pack(stream,END_CHILDREN_NODE)


def unpack_children(stream):
    children = []

    while True:
        node = Node.unpack(stream)

        if node.node_type == NodeType.END_GRAPH:
            raise FormatError('invalid scene graph')
        elif node.node_type == NodeType.BEGIN_CHILDREN:
            children[-1].children = unpack_children(stream)
        elif node.node_type == NodeType.END_CHILDREN:
            return children
        else:
            children.append(node)


def pack(stream,scene_graph):
    base = stream.tell()
    header = Header()
    header.unknown0 = scene_graph.unknown0
    header.shape_packet_count = scene_graph.shape_packet_count
    header.position_count = scene_graph.position_count
    stream.write(b'\x00'*Header.sizeof())

    header.node_offset = stream.tell() - base

    for node in scene_graph.children:
        Node.pack(stream,node)
        if node.children:
            Node.pack(stream,BEGIN_CHILDREN_NODE)
            pack_children(stream,node.children)

    Node.pack(stream,END_GRAPH_NODE)

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)


def unpack(stream):
    base = stream.tell()
    header = Header.unpack(stream)
    if header.magic != b'INF1':
        raise FormatError('invalid magic')

    scene_graph = SceneGraph()
    scene_graph.unknown0 = header.unknown0
    scene_graph.shape_packet_count = header.shape_packet_count
    scene_graph.position_count = header.position_count

    stream.seek(base + header.node_offset)

    while True:
        node = Node.unpack(stream)

        if node.node_type == NodeType.END_GRAPH:
            break
        elif node.node_type == NodeType.BEGIN_CHILDREN:
            scene_graph.children[-1].children = unpack_children(stream)
        elif node.node_type == NodeType.END_CHILDREN:
            raise FormatError('invalid scene graph')
        else:
            scene_graph.children.append(node)

    stream.seek(base + header.section_size)
    return scene_graph

#______________________________________________________________________________
