from btypes import *
import gx

#------------------------------------------------------------------------------

class Header(Struct):
    magic = ByteString(4)
    section_size = uint32
    entry_count = uint16
    __padding__ = Padding(2)
    shape_offset = uint32
    index_offset = uint32
    unknown0_offset = uint32
    vertex_descriptor_offset = uint32
    matrix_index_offset = uint32
    packet_offset = uint32
    matrix_selection_offset = uint32
    packet_location_offset = uint32

    def __init__(self):
        self.magic = b'SHP1'


class Shape(Struct):
    matrix_type = uint8
    __padding__ = Padding(1)
    batch_count = uint16
    vertex_descriptor_offset = uint16
    first_matrix_selection = uint16
    first_packet = uint16
    __padding__ = Padding(2)
    unknown0 = float32
    min_x = float32
    min_y = float32
    min_z = float32
    max_x = float32
    max_y = float32
    max_z = float32


class AttributeDescriptor(Struct):
    """Arguments to GXSetVtxDesc."""
    attribute = uint32
    input_type = uint32

    def __init__(self,attribute,input_type):
        self.attribute = attribute
        self.input_type = input_type


class MatrixSelection(Struct):
    unknown0 = uint16
    count = uint16
    first = uint32

    def __init__(self,first,count):
        self.unknown0 = 0
        self.first = first
        self.count = count


class PacketLocation(Struct):
    size = uint32
    offset = uint32

#------------------------------------------------------------------------------

def pack(stream,meshes):
    base = stream.tell()
    header = Header()
    header.entry_count = len(meshes)
    header.unknown0_offset = 0
    stream.write(b'\x00'*Header.sizeof())

    header.shape_offset = stream.tell() - base

    for i,mesh in enumerate(meshes):
        shape = Shape()

        shape.matrix_type = 0
        shape.batch_count = 1
        shape.first_matrix_selection = 0
        shape.first_packet = i
        shape.unknown0 = mesh.unknown0
        shape.min_x = mesh.min_x
        shape.min_y = mesh.min_y
        shape.min_z = mesh.min_z
        shape.max_x = mesh.max_x
        shape.max_y = mesh.max_y
        shape.max_z = mesh.max_z

        if not mesh.has_colors and not mesh.has_texcoords:
            shape.vertex_descriptor_offset = 0
        elif mesh.has_colors and not mesh.has_texcoords:
            shape.vertex_descriptor_offset = 16
        elif not mesh.has_colors and mesh.has_texcoords:
            shape.vertex_descriptor_offset = 40
        elif mesh.has_colors and mesh.has_texcoords:
            shape.vertex_descriptor_offset = 64

        Shape.pack(stream,shape)

    header.index_offset = stream.tell() - base
    for index in range(len(meshes)):
        uint16.pack(stream,index)

    # There's extra alignment here, probably because this is where the data at
    # unknown0_offset would be if there were any
    align(stream,4)

    align(stream,0x20)
    header.vertex_descriptor_offset = stream.tell() - base

    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_POS,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_NULL,gx.NONE))

    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_POS,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_CLR0,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_NULL,gx.NONE))

    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_POS,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_TEX0,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_NULL,gx.NONE))

    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_POS,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_CLR0,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_TEX0,gx.INDEX16))
    AttributeDescriptor.pack(stream,AttributeDescriptor(gx.VA_NULL,gx.NONE))

    header.matrix_index_offset = stream.tell() - base
    uint16.pack(stream,0)

    align(stream,0x20)
    header.packet_offset = stream.tell() - base

    packet_base = stream.tell()
    packet_locations = [PacketLocation() for _ in range(len(meshes))]

    for mesh,packet_location in zip(meshes,packet_locations):
        packet_location.offset = stream.tell() - packet_base

        for i,vertex in enumerate(mesh.vertices):
            if i % 0xFFFF == 0:
                uint8.pack(stream,gx.TRIANGLES)
                uint16.pack(stream,min(0xFFFF,len(mesh.vertices) - i))

            uint16.pack(stream,vertex.position_index)
            if mesh.has_colors:
                uint16.pack(stream,vertex.color_index)
            if mesh.has_texcoords:
                uint16.pack(stream,vertex.texcoord_index)

        align(stream,0x20,b'\x00')
        packet_location.size = (stream.tell() - packet_base) - packet_location.offset

    header.matrix_selection_offset = stream.tell() - base
    MatrixSelection.pack(stream,MatrixSelection(0,1))

    header.packet_location_offset = stream.tell() - base
    for packet_location in packet_locations:
        PacketLocation.pack(stream,packet_location)

    align(stream,0x20)
    header.section_size = stream.tell() - base
    stream.seek(base)
    Header.pack(stream,header)
    stream.seek(base + header.section_size)

#______________________________________________________________________________
