import sys
import os.path
import argparse
from btypes import *

import gx
import j3d
import j3d.inf1
import j3d.vtx1
import j3d.evp1
import j3d.drw1
import j3d.jnt1
import j3d.shp1
import j3d.mat3
import j3d.mdl3
import j3d.tex1

#------------------------------------------------------------------------------

class Position:

    def __init__(self,x,y,z):
        self.x = x
        self.y = y
        self.z = z


class Color:

    def __init__(self,r,g,b):
        self.r = r
        self.g = g
        self.b = b


class TextureCoordinate:

    def __init__(self,s,t):
        self.s = s
        self.t = t


class Vertex:

    def __init__(self,position_index,color_index,texcoord_index):
        self.position_index = position_index
        self.color_index = color_index
        self.texcoord_index = texcoord_index


class Mesh:

    def __init__(self,material_name):
        self.has_colors = False
        self.has_texcoords = False
        self.vertices = []
        self.min_x = None
        self.min_y = None
        self.min_z = None
        self.max_x = None
        self.max_y = None
        self.max_z = None
        self.material_name = material_name
        self.material_color = Color(0,0,0)
        self.texture_filename = None
        self.texture_index = None


# Parse arguments

parser = argparse.ArgumentParser(description='Convert wavefront OBJ files to Nintendo Wii BDL files')
parser.add_argument('input_filename',metavar='input.obj')
parser.add_argument('output_filename',nargs='?',metavar='output.bdl')
arguments = parser.parse_args()

if arguments.output_filename is None:
    arguments.output_filename = os.path.splitext(arguments.input_filename)[0] + '.bdl'

base_directory = os.path.dirname(arguments.input_filename)


# Read OBJ file

positions = []
colors = []
texcoords = []
mesh_dict = {}
current_mesh = None
material_library_filenames = set()

for line in open(arguments.input_filename):
    args = line.split()
    if len(args) == 0: continue
    command = args.pop(0)

    if command == 'mtllib':
        material_library_filenames.update(args)
    elif command == 'usemtl':
        current_mesh = mesh_dict.setdefault(args[0],Mesh(args[0]))
    elif command == 'v':
        x = float(args[0])
        y = float(args[1])
        z = float(args[2])
        positions.append(Position(x,y,z))
    elif command == 'vc':
        r = round(255*float(args[0]))
        g = round(255*float(args[1]))
        b = round(255*float(args[2]))
        colors.append(Color(r,g,b))
    elif command == 'vt':
        s = float(args[0])
        t = float(args[1])
        texcoords.append(TextureCoordinate(s,1 - t))
    elif command == 'f':
        for i in (0,2,1):
            indices = list(int(s) - 1 if s else None for s in args[i].split('/'))
            position_index = indices[0]
            texcoord_index = indices[1] if len(indices) > 1 else None
            color_index = indices[3] if len(indices) > 3 else None
            current_mesh.vertices.append(Vertex(position_index,color_index,texcoord_index))


# Read MTL files

for material_library_filename in material_library_filenames:
    current_mesh = None

    for line in open(os.path.join(base_directory,material_library_filename),'r'):
        args = line.split()
        if len(args) == 0: continue
        command = args.pop(0)

        if command == 'newmtl':
            current_mesh = mesh_dict.get(args[0])
        elif command == 'Kd' and current_mesh is not None:
            r = round(255*float(args[0]))
            g = round(255*float(args[1]))
            b = round(255*float(args[2]))
            current_mesh.material_color = Color(r,g,b)
        elif command == 'map_Kd' and current_mesh is not None:
            current_mesh.texture_filename = args[0]


meshes = []
textures = []

for mesh in mesh_dict.values():
    if not mesh.vertices:
        continue

    mesh.has_colors = all(v.color_index is not None for v in mesh.vertices)
    mesh.has_texcoords = all(v.texcoord_index is not None for v in mesh.vertices)
    mesh.min_x = min(positions[v.position_index].x for v in mesh.vertices)
    mesh.min_y = min(positions[v.position_index].y for v in mesh.vertices)
    mesh.min_z = min(positions[v.position_index].z for v in mesh.vertices)
    mesh.max_x = max(positions[v.position_index].x for v in mesh.vertices)
    mesh.max_y = max(positions[v.position_index].y for v in mesh.vertices)
    mesh.max_z = max(positions[v.position_index].z for v in mesh.vertices)
    # This isn't correct, but it works
    mesh.unknown0 = max(mesh.max_x - mesh.min_x,mesh.max_y - mesh.min_y,mesh.max_z - mesh.min_z)

    if mesh.texture_filename is not None and mesh.has_texcoords:
        mesh.texture_index = len(textures)
        textures.append(j3d.tex1.Texture(os.path.join(base_directory,mesh.texture_filename)))

    meshes.append(mesh)

if not meshes:
    raise Exception('model is empty')

root_joint = j3d.jnt1.Joint('root')
root_joint.min_x = min(mesh.min_x for mesh in meshes)
root_joint.min_y = min(mesh.min_y for mesh in meshes)
root_joint.min_z = min(mesh.min_z for mesh in meshes)
root_joint.max_x = max(mesh.max_x for mesh in meshes)
root_joint.max_y = max(mesh.max_y for mesh in meshes)
root_joint.max_z = max(mesh.max_z for mesh in meshes)
# This isn't correct, but it works
root_joint.unknown2 = max(root_joint.max_x - root_joint.min_x,root_joint.max_y - root_joint.min_y,root_joint.max_z - root_joint.min_z)

scene_graph = j3d.inf1.SceneGraph()
scene_graph.shape_packet_count = len(meshes)
scene_graph.position_count = len(positions)
scene_graph.children = [j3d.inf1.JointNode(0)]

current_parent = scene_graph.children[0]

for i,mesh in enumerate(meshes):
    current_parent.children = [j3d.inf1.MaterialNode(i)]
    current_parent.children[0].children = [j3d.inf1.ShapeNode(i)]
    current_parent = current_parent.children[0].children[0]


# Write BDL file

stream = FileStream(arguments.output_filename,'wb',BIG_ENDIAN)

header = j3d.Header()
header.magic = b'J3D2'
header.file_type = b'bdl4'
header.section_count = 9
header.unknown0 = b'SVR3'
stream.write(b'\x00'*j3d.Header.sizeof())

j3d.inf1.pack(stream,scene_graph)
j3d.vtx1.pack(stream,positions,colors,texcoords)
j3d.evp1.pack(stream)
j3d.drw1.pack(stream,[j3d.drw1.MatrixDescriptor(0,False)])
j3d.jnt1.pack(stream,[root_joint])
j3d.shp1.pack(stream,meshes)
j3d.mat3.pack(stream,meshes)
j3d.mdl3.pack(stream,meshes,textures)
j3d.tex1.pack(stream,textures)

header.file_size = stream.tell()
stream.seek(0)
j3d.Header.pack(stream,header)

#______________________________________________________________________________
